/*
   vol, a "simple volumes" for Linux
   Copyright (c) 1999  Martin Hinner <mhi@penguin.cz>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <linux/module.h>
#include <linux/malloc.h>
#include <linux/init.h>
#include <linux/genhd.h>
#include <linux/blkdev.h>
#include <linux/kernel.h>
#include <asm/uaccess.h>

#include "vol.h"

#ifdef DEBUG
#define dbg(x) {printk("vol:");printk x;}
#else
#define dbg(x)
#endif

#define MAJOR_NR  241

struct vol_info volumes[MAX_VOL_DEV];

extern int (*vol_map_ptr) (int, kdev_t *, unsigned long *,
			   unsigned long, int);

/*
 * block device support function for /usr/src/linux/drivers/block/ll_rw_blk.c
 */
static int
vol_map (int minor, kdev_t * rdev,
	 ulong * rsector, ulong size, int rw)
{
  ulong sec;
  struct vol_disk *disk;
  int diskno, i;

  dbg (("vol_map(minor=%u, rdev=%u.%u, rsector=%u, size=%u, rw=%u)\n",
	minor, MAJOR (*rdev), MINOR (*rdev), *rsector, size, rw));

  dbg(("blocksize = %u\n",vol_blocksizes[minor]));

  if (minor >= MAX_VOL_DEV)
    return -EFAULT;

  if (volumes[minor].ready == 0)
    return -EFAULT;

  switch (volumes[minor].type)
    {
    case VOL_LINEAR:
      disk = volumes[minor].disks;
      sec = *rsector;
      while (disk)
	{
//	  if (sec < disk->size / vol_blocksizes[minor])
	  if (sec < disk->size / 512)
	    break;
//	  sec -= disk->size / vol_blocksizes[minor];
	  sec -= disk->size / 512;
	  disk = disk->next;
	}

      if (!disk)		/* Sector not found: Abort, Retry, Fail? */
	return -EFAULT;

      *rsector = sec;
      *rdev = disk->dev;
      return 0;
    case VOL_STRIPE:
      diskno = ((*rsector * vol_blocksizes[minor]) / 0x10000)
	% volumes[minor].diskcount;

      i = 0;
      disk = volumes[minor].disks;
      while (disk)
	{
	  if (i == diskno)
	    break;
	  i++;
	  disk = disk->next;
	}

      if (!disk)		/* This should not happen... */
	return -EFAULT;

      *rsector =
	(((*rsector * vol_blocksizes[minor]) % 0x10000) /
	 vol_blocksizes[minor]) +

	(((*rsector * vol_blocksizes[minor]) /
	  (0x10000 * volumes[minor].diskcount)) *
	 (0x10000 / vol_blocksizes[minor]))
	;

      *rdev = disk->dev;
      dbg (("Stripe disk no %u,sector %u\n", diskno, *rsector));
      return 0;
    }
  return -EFAULT;		/* ... */
}

static int
vol_open (struct inode *inode, struct file *filp)
{
  int minor = MINOR (inode->i_rdev);

  if (minor >= MAX_VOL_DEV)
    return -EFAULT;

  volumes[minor].usecount++;
  dbg (("open\n"));
#ifdef MODULE
  MOD_INC_USE_COUNT;
#endif
  return 0;
}

static int
vol_close (struct inode *inode, struct file *filp)
{
  int minor = MINOR (inode->i_rdev);

  dbg (("close\n"));
  if (minor >= MAX_VOL_DEV)
    return -EFAULT;
  volumes[minor].usecount--;
#ifdef MODULE
  MOD_DEC_USE_COUNT;
#endif

  return 0;
}

static int 
vol_media_change (kdev_t dev)
{
  int rc;

  rc = vol_diskchanged[MINOR (dev)];
  vol_diskchanged[MINOR (dev)] = 0;

  return rc;
}

char *vol_names[3] =
{0, "Linear", "Stripe"};

static int
vol_ioctl (struct inode *inode, struct file *file,
	   uint command, ulong a)
{
  int minor = MINOR (inode->i_rdev);
  long *type;
  struct vol_add *vadd;
  struct vol_disk *vdisk, *vn;
  long size;

  if (minor >= MAX_VOL_DEV)
    return -EFAULT;

  dbg (("ioctl(command=%x)\n", command));
  switch (command)
    {
    case VOL_IOCTL_SETUP:
      type = (long *) a;
      dbg (("setting up disk vol%u,type=%u\n", minor, *type));
      if (volumes[minor].usecount > 1)
	return -EBUSY;
      volumes[minor].type = *type;
      vdisk = volumes[minor].disks;
      while (vdisk)
	{
	  vn = vdisk->next;
	  kfree (vdisk);
	  vdisk = vn;
	}
      volumes[minor].disks = 0;
      volumes[minor].diskcount = 0;
      volumes[minor].ready = 0;
      volumes[minor].size = 0;
      vol_sizes[minor] = 0;
      return 0;
    case VOL_IOCTL_ADD:
      vadd = (struct vol_add *) a;
      dbg (("adding device %u.%u (%s)\n",
	    MAJOR (vadd->dev), MINOR (vadd->dev), bdevname (vadd->dev)));

      vdisk = (struct vol_disk *) kmalloc (sizeof (struct vol_disk), GFP_KERNEL);
      if (vdisk == NULL)
	{
	  printk ("vol%u: kmalloc failed!\n", minor);
	  return -EFAULT;
	}
      vdisk->next = 0;
      vdisk->dev = vadd->dev;
      vdisk->size = vadd->size;
      vn = volumes[minor].disks;
      if (!vn)
	volumes[minor].disks = vdisk;
      else
	{
	  while (vn->next)
	    vn = vn->next;
	  vn->next = vdisk;
	}
      volumes[minor].diskcount++;
      switch (volumes[minor].type)
	{
	case VOL_LINEAR:
	  volumes[minor].size += vdisk->size;
	  break;
	case VOL_STRIPE:
	  if (volumes[minor].size)
	    {
	      if (volumes[minor].size > vdisk->size)
		volumes[minor].size = vdisk->size;
	    }
	  else
	    volumes[minor].size = vdisk->size;
	  break;
	}
      return 0;
    case VOL_IOCTL_START:
      if (volumes[minor].disks == NULL)
	return -EFAULT;
      volumes[minor].ready = 1;
      vol_sizes[minor] = volumes[minor].size / vol_blocksizes[minor];
      if (volumes[minor].type == VOL_STRIPE)
	volumes[minor].size *= volumes[minor].diskcount;
      vol_diskchanged[minor] = 1;

      printk ("vol%u: %s volume initialized, size = %u kB, %u disks.\n",
	      minor, vol_names[volumes[minor].type],
	      volumes[minor].size / 1024, volumes[minor].diskcount);

      return 0;
        case BLKGETSIZE:   /* Return device size */
                if (!a)
                        return -EINVAL;
                return put_user(vol_sizes[minor] << 1, (long *) a);
    default:
      return -EINVAL;
    }
  return -EINVAL;
}

static void vol_geninit (struct gendisk *ignored)
{
  dbg (("vol_geninit()\n"));
}

static struct file_operations vol_fops =
{
  NULL,				/* lseek - default */
  block_read,			/* read - general block-dev read */
  NULL, /*block_write,*/        /* write - general block-dev write */
  NULL,				/* readdir - bad */
  NULL,				/* poll */
  vol_ioctl,			/* ioctl */
  NULL,				/* mmap */
  vol_open,			/* open */
  NULL,				/* flush */
  vol_close,			/* release */
  0,				/* fsync */
  NULL,				/* fasync */
  vol_media_change,		/* media_change */
  0,				/* revalidate */
};

static struct hd_struct vol_hd_struct[MAX_VOL_DEV];
static int vol_sizes[MAX_VOL_DEV];
static int vol_blocksizes[MAX_VOL_DEV];
static int vol_diskchanged[MAX_VOL_DEV];

static struct gendisk vol_gendisk =
{
  MAJOR_NR,
  "vol",
  0,
  1,
  MAX_VOL_DEV,
  vol_geninit,
  vol_hd_struct,		/* partition table */
  vol_sizes,			/* vol sizes in blocks */
  MAX_VOL_DEV,			/* number of real devices */
  NULL,				/* internal */
  NULL				/* next disk */
};

void
vol_dummy_req (void)
{

}

int
register_vol (void)
{
  int i;

  for (i = 0; i < MAX_VOL_DEV; i++)
    {
      vol_gendisk.part[i].start_sect = -1;
      vol_sizes[i] = vol_gendisk.part[i].nr_sects = 0;
      vol_blocksizes[i] = 1024;
      vol_diskchanged[i] = 0;
      volumes[i].ready = 0;
      volumes[i].usecount = 0;
      volumes[i].type = VOL_UNUSED;
      volumes[i].disks = NULL;
      volumes[i].size = 0;
    }

  if (register_blkdev (MAJOR_NR, "vol", &vol_fops))
    {
      printk ("Unable to get major %d for vol\n", MAJOR_NR);
      return -EBUSY;
    }

  blk_dev[MAJOR_NR].request_fn = vol_dummy_req;
  blk_dev[MAJOR_NR].current_request = NULL;
  blk_size[MAJOR_NR] = vol_sizes;
  blksize_size[MAJOR_NR] = vol_blocksizes;


  vol_gendisk.next = gendisk_head;
  gendisk_head = &vol_gendisk;

  vol_map_ptr = vol_map;

  return 0;
}

void
unregister_vol (void)
{
  unregister_blkdev (MAJOR_NR, "vol");
}

#ifdef MODULE
int
init_module (void)
{
  printk ("Initializing VOL module\n");

  return (register_vol ());
}

void
cleanup_module (void)
{
  struct gendisk **gdp;

  unregister_vol ();

  blksize_size[MAJOR_NR] = NULL;
  blk_dev[MAJOR_NR].request_fn = NULL;
  blk_size[MAJOR_NR] = NULL;
  hardsect_size[MAJOR_NR] = NULL;
  read_ahead[MAJOR_NR] = 0;

  for (gdp = &gendisk_head; *gdp; gdp = &((*gdp)->next))
    if (*gdp == &vol_gendisk)
      break;
  if (*gdp)
    *gdp = (*gdp)->next;
  else
    dbg (("FATAL: vol_gendisk not found!\n"));
  vol_map_ptr = NULL;
  printk ("VOL module released\n");
}

#endif
